"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AuthProvider_1 = require("../AuthProvider");
const errors = require("../../errors");
const shared_1 = require("../../shared");
const crypto = require("crypto");
const moment = require("moment");
const PasswordRealm_1 = require("./PasswordRealm");
class PasswordAuthProvider extends AuthProvider_1.AuthProvider {
    constructor(config = {}) {
        super();
        this.name = 'password';
        this.autoCreateAdminUser = config.autoCreateAdminUser || false;
        this.saltLength = config.saltLength || 32;
        this.iterations = config.iterations || 10000;
        this.keyLength = config.keyLength || 512;
        this.digest = config.digest || "sha512";
        this.emailHandler = config.emailHandler;
    }
    start() {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            yield _super("start").call(this);
            this.passwordRealm = yield this.service.server.openRealm(PasswordRealm_1.PasswordRealm);
            if (this.autoCreateAdminUser) {
                try {
                    const username = "realm-admin";
                    const foundUser = this.service.getUserByProviderId(this.name, username);
                    if (!foundUser) {
                        const password = process.env.ROS_DEFAULT_PASSWORD || "";
                        yield this.attemptToRegister(username, password, true);
                        this.service.logger.info(`Autocreated admin user: ${username}`);
                    }
                }
                catch (err) {
                    this.service.logger.error('Failed to autocreate admin user', err);
                }
            }
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.passwordRealm) {
                this.passwordRealm.close();
            }
        });
    }
    authenticateOrCreateUser(body) {
        return __awaiter(this, void 0, void 0, function* () {
            let username;
            let password;
            let register;
            if (body["user_info"]) {
                username = body["data"];
                password = body["user_info"]["password"];
                register = body["user_info"]["register"];
            }
            else {
                username = body["username"];
                password = body["password"];
                register = body["register"];
            }
            if (!username) {
                throw new errors.realm.MissingParameters('username');
            }
            if (password === undefined || password === null) {
                throw new errors.realm.MissingParameters('password');
            }
            if (register === undefined) {
                const foundUser = this.service.getUserByProviderId('password', username);
                register = !foundUser;
            }
            if (register) {
                return this.attemptToRegister(username, password, false);
            }
            return this.attemptToLogin(username, password);
        });
    }
    update(user, data) {
        return __awaiter(this, void 0, void 0, function* () {
            const newPassword = data.new_password;
            if (!newPassword) {
                throw new errors.realm.MissingParameters('new_password');
            }
            yield this.updatePassword(newPassword, user.userId);
            return {};
        });
    }
    updateProviderAccount(providerId, user, data, isAuthenticated, userAgent, remoteIp) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.emailHandler) {
                throw new errors.realm.AccessDenied({
                    detail: 'The Password provider is not configured with an emailHandler.'
                });
            }
            switch (data.action) {
                case 'reset_password':
                    const token = crypto.randomBytes(32).toString('hex');
                    this.passwordRealm.beginTransaction();
                    if (user) {
                        this.passwordRealm.create('PasswordResetRequest', {
                            token,
                            expires: moment().add(12, 'hour').toDate(),
                            userId: user.userId
                        });
                        this.passwordRealm.commitTransaction();
                        this.emailHandler.resetPassword(providerId, token, userAgent, remoteIp)
                            .catch((err) => {
                            this.service.logger.error('Could not send reset email:', err);
                        });
                    }
                    else {
                        this.passwordRealm.cancelTransaction();
                    }
                    return {};
                case 'complete_reset':
                    const missingParams = [];
                    if (data.new_password === undefined || data.new_password === null) {
                        missingParams.push('new_password');
                    }
                    if (!data.token) {
                        missingParams.push('token');
                    }
                    if (missingParams.length > 0) {
                        throw new errors.realm.MissingParameters(...missingParams);
                    }
                    const existing = this.passwordRealm.objectForPrimaryKey('PasswordResetRequest', data.token);
                    if (!existing || existing.consumed || existing.expires < new Date()) {
                        throw new errors.realm.AccessDenied({
                            detail: 'The password reset token is invalid or has expired.'
                        });
                    }
                    this.passwordRealm.write(() => {
                        existing.consumed = new Date();
                    });
                    yield this.updatePassword(data.new_password, existing.userId);
                    return {};
                default:
                    throw new errors.realm.InvalidParameters({
                        name: 'action',
                        reason: `The action ${data.action} is not supported.`
                    });
            }
        });
    }
    deleteUser(userId) {
        return __awaiter(this, void 0, void 0, function* () {
            let userDeleted = false;
            this.passwordRealm.write(() => {
                const foundPassword = this.passwordRealm.objectForPrimaryKey('PasswordSaltHash', userId);
                if (foundPassword) {
                    userDeleted = true;
                    this.passwordRealm.delete(foundPassword);
                }
            });
            return userDeleted;
        });
    }
    updatePassword(newPassword, userId) {
        return __awaiter(this, void 0, void 0, function* () {
            const salt = yield this.createSalt(this.saltLength);
            const hash = yield this.hashPassword(newPassword, salt, this.iterations, this.keyLength, this.digest);
            this.passwordRealm.write(() => {
                this.passwordRealm.create('PasswordSaltHash', {
                    salt: salt.toString("base64"),
                    iterations: this.iterations,
                    keyLength: this.keyLength,
                    digest: this.digest,
                    hash: hash.toString("base64"),
                    userId: userId,
                }, true);
            });
        });
    }
    hashPassword(clearTextPassword, salt, iterations, keyLength, digest) {
        return __awaiter(this, void 0, void 0, function* () {
            return shared_1.Promisify(crypto.pbkdf2.bind(crypto), clearTextPassword, salt, iterations, keyLength, digest);
        });
    }
    createSalt(saltLength) {
        return __awaiter(this, void 0, void 0, function* () {
            return shared_1.Promisify(crypto.randomBytes.bind(crypto), saltLength);
        });
    }
    comparePassword(candidatePassword, salt, iterations, keyLength, digest, hash) {
        return __awaiter(this, void 0, void 0, function* () {
            const hashedCandidatePassword = yield this.hashPassword(candidatePassword, salt, iterations, keyLength, digest);
            return hashedCandidatePassword.equals(hash);
        });
    }
    attemptToRegister(username, password, isAdmin) {
        return __awaiter(this, void 0, void 0, function* () {
            const foundUser = this.service.getUserByProviderId(this.name, username);
            if (foundUser) {
                throw new errors.realm.InvalidCredentials();
            }
            const user = yield this.service.createOrUpdateUser(username, this.name, isAdmin);
            const userId = user.userId;
            const salt = yield this.createSalt(this.saltLength);
            const hash = yield this.hashPassword(password, salt, this.iterations, this.keyLength, this.digest);
            this.passwordRealm.write(() => {
                this.passwordRealm.create('PasswordSaltHash', {
                    salt: salt.toString("base64"),
                    iterations: this.iterations,
                    keyLength: this.keyLength,
                    digest: this.digest,
                    hash: hash.toString("base64"),
                    userId: userId
                }, true);
            });
            return user;
        });
    }
    attemptToLogin(username, password) {
        return __awaiter(this, void 0, void 0, function* () {
            const foundUser = this.service.getUserByProviderId(this.name, username);
            if (!foundUser) {
                const salt = yield this.createSalt(this.saltLength);
                const hash = yield this.hashPassword(password, salt, this.iterations, this.keyLength, this.digest);
                throw new errors.realm.InvalidCredentials();
            }
            const foundPassword = this.passwordRealm.objectForPrimaryKey('PasswordSaltHash', foundUser.userId);
            if (!foundPassword) {
                const salt = yield this.createSalt(this.saltLength);
                const hash = yield this.hashPassword(password, salt, this.iterations, this.keyLength, this.digest);
                throw new errors.realm.InvalidCredentials();
            }
            const passwordMatch = yield this.comparePassword(password, new Buffer(foundPassword.salt, "base64"), foundPassword.iterations, foundPassword.keyLength, foundPassword.digest, new Buffer(foundPassword.hash, "base64"));
            if (!passwordMatch) {
                throw new errors.realm.InvalidCredentials();
            }
            return foundUser;
        });
    }
}
exports.PasswordAuthProvider = PasswordAuthProvider;
//# sourceMappingURL=PasswordAuthProvider.js.map